﻿/*** Name:      Game.cs
 *** Created:   11/6/2008, by: Hilkens, RGM
 *** Modified:  12/14/2008, by: Hilkens, RGM
 *** Source:    Hoofdklasse van het spel Etaris. Deze klasse bevat de spellogica.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using Etaris.Fields;
using System.Security.Cryptography;

namespace Etaris.Core
{
    /// <summary>
    /// Enumeratie van unieke speelvelden per spel.
    /// </summary>
    public enum FieldTypes
    {
        /// <summary>
        /// Bovenste speelveld.
        /// </summary>
        TopField = 0,

        /// <summary>
        /// Onderste speelveld.
        /// </summary>
        BottomField = 1
    }

    /// <summary>
    /// Hoofdklasse van het spel Etaris. Deze klasse bevat de spellogica.
    /// </summary>
    /// <remarks>
    /// Deze klasse definieert een set van speelstuk-kleuren.
    /// </remarks>
    public class Game
    {
        #region "Delegates"
        /// <summary>
        /// Delegate voor GameOver gebeurtenis.
        /// </summary>
        /// <param name="score">Eindscore.</param>
        public delegate void GameOverHandler(int score);

        /// <summary>
        /// Delegate voor PreviewNextPiece gebeurtenis.
        /// </summary>
        /// <param name="template">Een speelstuktemplate.</param>
        /// <param name="fieldType">Speelveldtype behorende bij het speelstuktemplate.</param>
        public delegate void PreviewNextPieceHandler(PieceTemplate template, FieldTypes fieldType);
        #endregion

        #region "Handlers"
        /// <summary>
        /// Dit is een gebeurtenis om bovenliggende klassen te laten weten dat het spel game-over is.
        /// </summary>
        public event GameOverHandler GameOver;

        /// <summary>
        /// Dit is een gebeurtenis om bovenliggende klassen een volgend speelstuktemplate voor een specifiek speelveld te laten weten.
        /// </summary>
        public event PreviewNextPieceHandler PreviewNextPiece;
        #endregion

        #region Velden
        /// <summary>
        /// De twee speelvelden.
        /// </summary>
        PlayingField[] fields = new PlayingField[2];

        /// <summary>
        /// Het actieve speelveld.
        /// </summary>
        FieldTypes activeField = FieldTypes.TopField;

        /// <summary>
        /// Het huidige configuratiebestand.
        /// </summary>
        ConfigurationFile config;

        /// <summary>
        /// Algemene timer voor het 'vallen' van speelstukken. Deze klasse verwerkt hiervan niet direct gebeurtenissen.
        /// </summary>
        Timer timer;

        /// <summary>
        /// Een random-getallen generator.
        /// </summary>
        RNGCryptoServiceProvider[] rng = new RNGCryptoServiceProvider[2];

        /// <summary>
        /// Volgende speelstukken voor beide speelvelden.
        /// </summary>
        PieceTemplate[] nextTemplates = new PieceTemplate[2]; //indexen komen overeen met FieldTypes.

        /// <summary>
        /// De kleuren die speelstukken aan kunnen nemen.
        /// </summary>
        Color[] colors = new Color[] { Color.Red, Color.Blue, Color.Yellow };

        /// <summary>
        /// Het actieve speelstuk, dit kan door <i>TimerTick</i> en <i>ProcessKey</i> bewogen worden.
        /// </summary>
        Piece activePiece;

        /// <summary>
        /// De huidige score.
        /// </summary>
        int score = 0;

        /// <summary>
        /// De hoogte van het bovenste (Top > 0) gedeelte van het speelveld dat niet tot het daadwerkelijke speelveld behoord. Nieuwe speelstukken bevinden zich intitieel in dit gebied.
        /// </summary>
        int bottomOffset;

        /// <summary>
        /// Het (begin)niveau dat de gebruiker heeft gekozen.
        /// </summary>
        int level;

        /// <summary>
        /// De versnelling die de blokken gedurende het hele spel hebben.
        /// </summary>
        double speedupRate;

        /// <summary>
        /// De tijd dat de gebruiker al bezig is met dit spel, in milliseconden.
        /// </summary>
        ulong playingTime;

        /// <summary>
        /// De huidige snelheid.
        /// </summary>
        double gameSpeed;

        /// <summary>
        /// De basissnelheid. De beginsnelheid is <i>level</i> * <i>defaultSpeed</i>.
        /// </summary>
        const int defaultSpeed = 30;

        #endregion

        #region "Properties"
        /// <summary>
        /// Bovenste speelveld.
        /// </summary>
        public PlayingField TopField
        {
            get { return fields[(int)FieldTypes.TopField]; }
            private set { fields[(int)FieldTypes.TopField] = value; }
        }

        /// <summary>
        /// Onderste speelveld.
        /// </summary>
        public PlayingField BottomField
        {
            get { return fields[(int)FieldTypes.BottomField]; }
            private set { fields[(int)FieldTypes.BottomField] = value; }
        }

        /// <summary>
        /// Het actieve speelveld.
        /// </summary>
        public FieldTypes ActiveField
        {
            get { return activeField; }
        }

        /// <summary>
        /// Huidige score.
        /// </summary>
        public int Score
        {
            get { return score; }
        }
        #endregion

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        /// <param name="configurationFile">Een configuratiebestand voor het nieuwe spel.</param>
        /// <param name="timer">Algemene timer voor het 'vallen' van speelstukken. Deze klasse verwerkt hiervan niet direct gebeurtenissen.</param>
        /// <param name="level">Het gekozen niveau.</param>
        /// <param name="speedupRate">De gekozen speed-up rate.</param>
        /// <param name="onGameLoading">Handler welke uitgevoerd wordt vlak voordat het spel gestart wordt.</param>
        public Game(ConfigurationFile configurationFile, Timer timer, int level, double speedupRate, EventHandler onGameLoading)
        {
            //Init.
            int temp = 0;
            Settings settings = Settings.CurrentSettings;

            //Stel waardes in.
            this.config = configurationFile;
            this.timer = timer;
            this.config.GetMaximumPieceDimensions(out temp, out bottomOffset);
            this.speedupRate = speedupRate;
            this.level = level;
            TopField = new PlayingField(settings.BlockSize, settings.ShadingSize, config.PlayingFieldWidth, config.PlayingFieldHeight + bottomOffset);
            BottomField = new PlayingField(settings.BlockSize, settings.ShadingSize, config.PlayingFieldWidth, config.PlayingFieldHeight + bottomOffset);
            rng[0] = new RNGCryptoServiceProvider();
            rng[1] = new RNGCryptoServiceProvider();
            activeField = FieldTypes.TopField;

            //Instellen valsnelheid.
            gameSpeed = level * defaultSpeed;
            timer.Interval = (int)(60000 / gameSpeed);

            //Event versturen en spel activeren.
            onGameLoading.Invoke(this, null);
            GetNextPieceTemplate(FieldTypes.BottomField);
            GetNextPieceTemplate(FieldTypes.TopField);
            InsertRandomPiece();
            activePiece.Move(Direction.Up);
        }

        /// <summary>
        /// <i>true</i> als het spel gepauzeerd is, anderzijds <i>false</i>.
        /// </summary>
        public bool Paused
        {
            get { return timer.Enabled; }
            set
            { timer.Enabled = !value; }
        }

        /// <summary>
        /// Verwerkt een tik van de timer. De positie van een speelstuk wordt indien nodig veranderd.
        /// </summary>
        public void TimerTick()
        {
            if (activePiece == null)
            {
                if (DropAllFloatingPieces()) //Als er zwevende stukken waren
                {
                    Speed();
                    return;
                }
                else if (RemoveFullRows() > 0)
                {
                    Speed();
                    return;
                }

                //Wissel van speelveld en maak een nieuw stuk als de gebruiker niet game-over is.
                if (fields[(int)activeField].IsRowEmpty(config.PlayingFieldHeight))
                {
                    activeField = activeField == FieldTypes.BottomField ? FieldTypes.TopField : FieldTypes.BottomField;
                    InsertRandomPiece();
                    GetNextPieceTemplate(activeField);
                }
                else
                {
                    Paused = true;
                    if (GameOver != null) GameOver(score);
                }
            }

            //Beweeg het actieve speelstuk omlaag
            if ((activePiece != null) && (!activePiece.Move(Direction.Up)))
            {
                //Indien dat niet kan:
                activePiece = null;
                TimerTick(); return; //Zoek uit wat er nu moet gebeuren...
            }
            Speed();
        }

        /// <summary>
        /// Verwerkt gebruikersinvoer van het toetsenbord.
        /// </summary>
        /// <param name="key">Een toetsaanslag.</param>
        public void ProcessKey(Keys key)
        {
            if (activePiece==null) return; //Als er geen actief speelstuk is, niks doen met beweging
            switch (key)
            {
                case Keys.Left: activePiece.Move(Direction.Left);
                    break;
                case Keys.Right: activePiece.Move(Direction.Right);
                    break;
                case Keys.Down: activePiece.Move(Direction.Up);
                    break;
                case Keys.ShiftKey: activePiece.Rotate();
                    break;
            }
        }

        /// <summary>
        /// Berekent de nieuwe snelheid.
        /// </summary>
        private void Speed()
        {
            //playingtime in miliseconden
            playingTime = playingTime + (ulong)timer.Interval;

            //Ticks per 60 seconden
            gameSpeed = level * defaultSpeed + ((speedupRate * playingTime) / 60000);
            timer.Interval = (int)(60000 / gameSpeed);

        }

        /// <summary>
        /// Bepaalt een volgend speelstuktemplate en slaat het nieuwe template op in nextTemplates.
        /// </summary>
        /// <param name="fieldType">Speelveldtype waarvoor het nieuwe speelstuktemplate bepaald moet worden.</param>
        private void GetNextPieceTemplate(FieldTypes fieldType)
        {
            //Init
            byte[] buffer = new byte[1];

            int next = GetRandom(config.PieceTemplatesCount - 1);   //Volgende speelstuk.
            int nextColor = GetRandom(colors.Length - 1);           //Volgende kleur-index.
            Color color;                                            //Volgende kleur.

            //Controle op uniek speelstuk (als er maar één template is, dan wordt alleen dat template gebruikt).
            nextTemplates[(int)fieldType] = config.GetPieceTemplate(next).Clone();

            //Bepaal kleur.
            color = colors[nextColor];
            nextTemplates[(int)fieldType].Color = color;

            //Verstuur event.
            if (PreviewNextPiece != null) PreviewNextPiece(nextTemplates[(int)fieldType], fieldType);
        }

        /// <summary>
        /// Voegt een nieuw speelstuk op basis van een eerder bepaald speelstuktemplate toe aan het actieve speelveld.
        /// </summary>
        private void InsertRandomPiece()
        {
            //Init.
            PieceTemplate template = nextTemplates[(int)activeField];
            Piece piece;
            List<BlockTemplate> blocks = template.Blocks;
            int xOffset, yOffset;
            int fieldWidth = config.PlayingFieldWidth;
            int fieldHeight = config.PlayingFieldHeight;

            //Maak speelstuk.
            piece = new Piece(fields[(int)activeField], template.Color);
 
            //Bereken offsets.
            xOffset = (int)((fieldWidth / 2.0) - (template.Width / 2.0));
            yOffset = fieldHeight;

            //Voorkom incorrecte waardes door afrondingsfouten.
            if (xOffset < 0) xOffset = 0;
            if ((xOffset + template.Width) > fieldWidth) xOffset = fieldWidth - template.Width - 1;

            //Voeg blokken toe aan speelstuk.
            for (int i = 0; i < template.Blocks.Count; i++)
                piece.AddBlock(blocks[i].Left + xOffset, blocks[i].Top + yOffset);

            activePiece = piece;
        }

        /// <summary>
        /// Laat alle speelstukken in het huidige speelveld maximaal vallen door de zwaartekracht en verwijder nieuwe volle rijen.
        /// </summary>
        /// <comment>Deze methode wordt momenteel niet gebruikt.</comment>
        private void MaximalDropAllPieces()
        {
            //Init.
            bool run = true;
            LinkedList<Piece> pieces = fields[(int)activeField].Pieces;

            //Laat alle speelstukken zo ver mogelijk vallen, en verwijder nieuwe volle rijen, zodra de stukken niet meer vallen
            while (run)
            {
                //Laat alle speelstukken zo ver mogelijk vallen
                while (run)
                {
                    run = false;
                    foreach (Piece piece in pieces)
                        run = run || piece.MoveMaximal(Direction.Up);
                }
                //Verwijder nieuwe volle rijen
                if (RemoveFullRows() > 0)
                    run = true;
            }
        }

        /// <summary>
        /// Laat alle zwevende speelstukken in het actieve speelveld een rij vallen door de zwaartekracht.
        /// </summary>
        /// <returns><i>true</i> als er een zwevend speelstuk was, anderzijds <i>false</i>.</returns>
        private bool DropAllFloatingPieces()
        {
            //Init.
            bool run = true, ret = false;
            int i;
            List<Piece> pieces = new List<Piece>();

            //Maak een kopie van alle speelstukken, zodat ze uit deze lijst kunnen worden verwijderd
            foreach (Piece piece in fields[(int)activeField].Pieces)
                pieces.Add(piece);

            //Doorloop alle speelstukken tot een stuk al gevallen is, of niet meer kan vallen.
            while (run)
            {
                i = 0; //Begin (opnieuw) over alle stukken te itereren.
                run = false;
                while (i < pieces.Count)
                    if (pieces[i].Move(Direction.Up))
                    {
                        pieces.Remove(pieces[i]);
                        ret = true;
                        run = true;
                    }
                    else
                        i++;
            }
            return ret;
        }
        
        /// <summary>
        /// Verwijdert alle volle rijen en werkt de score bij.
        /// </summary>
        /// <returns>Het aantal verwijderde rijen.</returns>
        private int RemoveFullRows()
        {
            //Init.
            PlayingField field = fields[(int)activeField];
            List<int> rows = new List<int>();
            int[] rowsArray;

            //Deze implementatie is O(n), maar de daadwerkelijke operaties zijn zéér snel. Dit is intern afgesproken!
            for (int i = 0; i < config.PlayingFieldHeight; i++)
            {
                if (field.IsRowFull(i)) rows.Add(i);
            }

            //Werk de score bij.
            rowsArray = new int[rows.Count];
            for (int i = 0; i < rows.Count; i++) rowsArray[i] = rows[i];
            if (rows.Count > 0) CalculateScore(rowsArray);

            //Verwijder de rijen.
            field.RemoveRows(rowsArray);

            //Finish.
            return rows.Count;
        }

        /// <summary>
        /// Berekent de nieuwe score als de rijen <i>rows</i> in het actieve speelveld verwijderd worden.
        /// </summary>
        /// <param name="rows">Array met de te verwijderen rijnummers.</param>
        private void CalculateScore(int[] rows)
        {
            //Init.
            double totalPartial = 0;
            const double alpha = 10;
            const double beta = 0.01;
            const double gamma = 1;

            //Berekent de volledige gedeeltelijke score.
            for (int i = 0; i < rows.Length; i++)
                totalPartial += CalculatePartialScore(rows[i]);

            //Bereken de nieuwe eindscore.
            score += (int)(alpha * ((double)level + beta * (double)playingTime / 1000 * speedupRate) * (-gamma + totalPartial)); //playingTime is in milliseconden, en niet in seconden.
        }

        /// <summary>
        /// Berekent de gedeeltelijke nieuwe score als de rij <i>row</i> in het actieve speelveld vol is. Deze functie wordt aangeroepen voordat de betreffende rij verwijderd wordt.
        /// </summary>
        /// <param name="row">Rijnummer in het actieve speelveld.</param>
        /// <returns>De gedeeltelijke score.</returns>
        private double CalculatePartialScore(int row)
        {
            //Init.
            const double delta = 2;
            PlayingField field = fields[(int)activeField];
            Color color;
            bool[] readColors = new bool[colors.Length];
            int differentColors = 0;
            int colorIndex;
            int i, j;

            //Zet readColors expliciet op false.
            for (i = 0; i < colors.Length; i++) readColors[i] = false;

            //Doorloop alle blokken.
            for (i = 0; i < config.PlayingFieldWidth; i++)
            {
                color = field.GetPiece(i, row).Color;

                for (j = 0; j < colors.Length; j++)
                    if (colors[j] == color) { colorIndex = j; break; }

                readColors[j] = true;
                if (j == colors.Length) throw new Exception("Illegal color.");
            }

            //Bepaal het aantal verschillende kleuren.
            for (i = 0; i < readColors.Length; i++)
                if (readColors[i] == true) differentColors += 1;

            //Finish.
            return 3 + delta - differentColors;
        }

        /// <summary>
        /// Genereert een willekeurig getal tussen 0 en <i>maximum</i>.
        /// </summary>
        /// <param name="maximum">De maximale waarde van het willekeurige getal.</param>
        /// <returns>Een willekeurig getal.</returns>
        public int GetRandom(int maximum)
        {
            //Init.
            byte[] buffer = new byte[4];
            int value;
            double dblvalue;
            int temp;

            rng[(int)activeField].GetBytes(buffer);
            value = buffer[0]; value += buffer[1] << 8;
            value += buffer[2] << 16; value += buffer[3] << 24;

            //return Math.Abs(value % (maximum + 1)); //dit is toch beter

            //dblvalue is random en heeft een waarde tussen 0 en maximum + 1.
            dblvalue = (Math.Abs(value) / (double)(int.MaxValue)) * (maximum + 1);

            /* Wanneer dbl wordt uitgerekend tussen 0 en maximum gaat het fout.
            *  Immers, 
            *  De waarde 0 kan verkregen worden als naar beneden wordt afgerond.
            *  De waarde maximum kan allen verkregen worden als naar boven wordt afgerond.
            *  Andere getallen kunnen verkregen wordt als zowel naar boven als naar beneden wordt afgerond.
            *  Door het bereik op [0, maximum + 1] te kiezen, kan een waarde welke naar naar boven afrond naar maximum + 1,
            *  omgezet worden naar de waarde 0. De waarde maximum kun nu ook verkregen worden door naar beneden af te ronden.
            *  Daar afronding met / en * is de % methode wellicht beter.
            */

            temp = (int)Math.Round(dblvalue);
            if (temp > maximum) return 0; else return temp; 
        }
    }
}
