﻿/*** Name:      PieceTemplate.cs
 *** Created:   11/18/2008, by: Haan, S de
 *** Modified:  12/14/2008, by: Hilkens, RGM
 *** Source:    Definieert een template voor een speelstuk.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using Etaris.Fields;
using System.Drawing;

namespace Etaris.Core
{
    /// <summary>
    /// Definieert een speelstuktemplate.
    /// </summary>
    public class PieceTemplate : BasePiece<BlockTemplate>
    {
        #region Velden
        /// <summary>
        /// Breedte van het speelstuktemplate in blokken.
        /// </summary>
        int width = 0;

        /// <summary>
        /// Hoogte van het speelstuktemplate in blokken.
        /// </summary>
        int height = 0;

        /// <summary>
        /// X-positie van het meest linkse blok.
        /// </summary>
        int minLeft = 0;

        /// <summary>
        /// X-positie van het meest rechtse blok.
        /// </summary>
        int maxLeft = 0;

        /// <summary>
        /// Y-positie positie van het hoogste blok.
        /// </summary>
        int minTop = 0;

        /// <summary>
        /// Y-positie van het laagste blok.
        /// </summary>
        int maxTop = 0;
        #endregion

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        internal PieceTemplate()
            : base(Color.Red)
        { }

        /// <summary>
        /// Kloont het huidige object.
        /// </summary>
        /// <returns>Een kopie van dit object.</returns>
        public PieceTemplate Clone()
        {
            //Init.
            PieceTemplate newP = (PieceTemplate)MemberwiseClone();

            newP.blocks = new List<BlockTemplate>();
            foreach (BlockTemplate block in blocks)
                newP.AddBlock(block.Left, block.Top);

            //Finish.
            return newP;
        }

        /// <summary>
        /// Breedte van het speelstuktemplate in blokken.
        /// </summary>
        public int Width
        {
            get { return width; }
        }

        /// <summary>
        /// Hoogte van het speelstuktemplate in blokken.
        /// </summary>
        public int Height
        {
            get { return height; }
        }

        /// <summary>
        /// Lijst met blokken die horen bij dit speelstuk.
        /// </summary>
        internal List<BlockTemplate> Blocks
        {
            get { return blocks; }
        }

        /// <summary>
        /// Voegt een blok toe (houdt geen rekening met consistentie).
        /// </summary>
        /// <param name="left">Kolom in speelveld (beginnend vanaf 0).</param>
        /// <param name="top">Rij in speelveld (beginnend vanaf 0).</param>
        public void AddBlock(int left, int top)
        {
            Blocks.Add(new BlockTemplate(left, top));
        }

        /// <summary>
        /// Controleert of het speelstuktemplate voldoet aan de gestelde eisen.
        /// </summary>
        /// <returns><i>true</i> als het template geldig is, anderzijds <i>false</i>.</returns>
        internal bool ValidatePiece()
        {
            //Init.
            int countPart = 0;
            bool[,] matrix;

            //Bereken dimensies en controleer template.
            CalculateDimensions();

            //Transformeer speelstukken door verwijderen lege ruimte.
            for (int i = 0; i < blocks.Count; i++)
            {
                blocks[i].Left = blocks[i].Left - minLeft;
                blocks[i].Top = blocks[i].Top - minTop;
            }

            //Laad blokken in de matrix.
            matrix = new bool[width, height];
            for (int i = 0; i < blocks.Count; i++)
                matrix[blocks[i].Left, blocks[i].Top] = true;

            CountPart(ref matrix, blocks[0].Left, blocks[0].Top, ref countPart);

            //Finish.
            return (countPart == blocks.Count);
        }

        /// <summary>
        /// Berekent een kleinste omsloten rechthoek voor dit speelstuktemplate. 
        /// </summary>
        private void CalculateDimensions()
        {
            //Controleer Blocks.
            if (blocks.Count == 0)
            { minTop = 0; minTop = 0; maxTop = 0; maxLeft = 0; width = 0; height = 0; return; }

            minLeft = blocks[0].Left;
            minTop = blocks[0].Top;

            maxLeft = minLeft;
            maxTop = minTop;

            //Doorloop alle blokken.
            for (int i = 1; i < blocks.Count; i++)
            {
                if (blocks[i].Left < minLeft)
                    minLeft = blocks[i].Left;
                if (blocks[i].Left > maxLeft)
                    maxLeft = blocks[i].Left;
                if (blocks[i].Top < minTop)
                    minTop = blocks[i].Top;
                if (blocks[i].Top > maxTop)
                    maxTop = blocks[i].Top;
            }

            //Finish.
            width = maxLeft - minLeft + 1;
            height = maxTop - minTop + 1;
        }

        /// <summary>
        /// Telt alle verbonden blokken in <i>matrix</i> (consistentie controle).
        /// </summary>
        /// <param name="matrix">Matrix met booleans. <i>true</i> geeft een niet verwerkt blok aan. <i>false</i> geeft een verwerkt of een niet bestaand blok aan.</param>
        /// <param name="left">Kolom in matrix van startpositie (positie in matrix met waarde <i>true</i>).</param>
        /// <param name="top">Rij in matrix van startpositie (positie in matrix met waarde <i>true</i>).</param>
        /// <param name="count">Aantal gevonden blokken.</param>
        private void CountPart(ref bool[,] matrix, int left, int top, ref int count)
        {
            matrix[left, top] = false;
            count++;

            if (left > 0)
            {
                if (matrix[left - 1, top] == true)
                    CountPart(ref matrix, left - 1, top, ref count);
            }

            if ((left + 1) < width)
            {
                if (matrix[left + 1, top] == true)
                    CountPart(ref matrix, left + 1, top, ref count);
            }

            if (top > 0)
            {
                if (matrix[left, top - 1] == true)
                    CountPart(ref matrix, left, top - 1, ref count);
            }

            if ((top + 1) < height)
            {
                if (matrix[left, top + 1] == true)
                    CountPart(ref matrix, left, top + 1, ref count);
            }
        }
    }
}