﻿/*** Name:      BasePiece.cs
 *** Created:   11/21/2008, by: Hilkens, RGM
 *** Modified:  12/6/2008, by: Hilkens, RGM
 *** Source:    Definieert een basisspeelstuk.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using sysui;

namespace Etaris.Fields
{
    /// <summary>
    /// Definieert een basisspeelstuk. Deze klasse implementeert het tekenen van speelstukken.
    /// </summary>
    /// <typeparam name="T">Het type blokken waaruit het speelstuk bestaat</typeparam>
    public abstract class BasePiece<T> where T : BaseBlock
    {
        #region Velden
        /// <summary>
        /// Hoofdkleur van dit speelstuk
        /// </summary>
        private Color color;

        /// <summary>
        /// Lijst met blokken van dit speelstuk.
        /// </summary>
        protected List<T> blocks;

        /// <summary>
        /// Schaduwkleur van dit speelstuk
        /// </summary>
        private Color shadingColor = Color.White;

        /// <summary>
        /// Set van pennen
        /// </summary>
        private Pen[] pens;

        /// <summary>
        /// Lengte/breedte van een blok
        /// </summary>
        private int blockSize;

        /// <summary>
        /// Diepte van een gradient (schaduwrand).
        /// </summary>
        private int shadingSize;
        #endregion

        /// <summary>
        /// Hoofdkleur van dit speelstuk.
        /// </summary>
        public Color Color
        {
            get { return color; }
            set
            {
                //Kleur opslaan.
                color = value;

                //Bepaal shading kleur.
                if (color == Color.Blue) shadingColor = Color.DarkBlue;
                else if (color == Color.Yellow) shadingColor = Color.Orange;
                else if (color == Color.Red) shadingColor = Color.DarkRed;

                //Bepaal 'colors'.
                pens = ComputeGradientColors(shadingColor, color, shadingSize);
            }
        }

        /// <summary>
        /// Lijst met blokken van dit speelstuk.
        /// </summary>
        internal List<T> Blocks
        {
            get { return blocks; }
        }

        /// <summary>
        /// Stelt de afmetingen <i>blockSize</i> en <i>shadingSize</i> in.
        /// </summary>
        /// <param name="blockSize">Lengte/breedte van een blok.</param>
        /// <param name="shadingSize">Diepte van een gradient (schaduwrand).</param>
        public void SetPaintingSizes(int blockSize, int shadingSize)
        {
            if (blockSize < 1 || shadingSize < 1) return;
            this.blockSize = blockSize;
            this.shadingSize = shadingSize;
            pens = ComputeGradientColors(shadingColor, color, shadingSize);
        }

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        /// <param name="color">Hoofdkleur van dit speelstuk.</param>
        protected BasePiece(Color color)
        {
            //Init.
            blocks = new List<T>();
            SetPaintingSizes(15, 6);    //Standaard afmetingen.
            this.Color = color;         //Stel ook de schaduwkleur in!
        }

        /// <summary>
        /// Tekent het speelstuk.
        /// </summary>
        /// <remarks>De prestatie van deze functie is 30% tot 40% beter dan de vorige implementatie.</remarks>
        /// <param name="g">Een <i>Graphics</i> object.</param>
        public void Paint(Graphics g)
        {
            //Init.
            int i, j;
            Border<T> border = new Border<T>();
            List<Point[][]> points = border.ProcessPiece(this, blockSize, shadingSize);
            Point[] blockPoints = new Point[4];

            //Vul speelstuk.
            foreach (BaseBlock block in Blocks)
            {
                blockPoints[0] = new Point(block.Left * blockSize, block.Top * blockSize);
                blockPoints[1] = new Point((block.Left + 1) * blockSize - 1, block.Top * blockSize);
                blockPoints[2] = new Point((block.Left + 1) * blockSize - 1, (block.Top + 1) * blockSize - 1);
                blockPoints[3] = new Point(block.Left * blockSize, (block.Top + 1) * blockSize - 1);

                g.DrawPolygon(new Pen(color), blockPoints);
                g.FillPolygon(new SolidBrush(color), blockPoints);
            }

            //Teken alle borders
            for (i = 0; i < points.Count; i++)
            {
                for (j = 0; j < shadingSize; j++) g.DrawPolygon(pens[j], points[i][j]);
            }
        }

        /// <summary>
        /// Wist de afbeelding van het scherm.
        /// </summary>
        /// <param name="g">Een <i>Graphics</i> object.</param>
        public void ClearDrawing(Graphics g)
        {

            //Init.
            Point[] blockPoints = new Point[4];

            //Vul speelstuk.
            foreach (BaseBlock block in Blocks)
            {
                blockPoints[0] = new Point(block.Left * blockSize, block.Top * blockSize);
                blockPoints[1] = new Point((block.Left + 1) * blockSize - 1, block.Top * blockSize);
                blockPoints[2] = new Point((block.Left + 1) * blockSize - 1, (block.Top + 1) * blockSize - 1);
                blockPoints[3] = new Point(block.Left * blockSize, (block.Top + 1) * blockSize - 1);

                g.DrawPolygon(new Pen(Color.White), blockPoints);
                g.FillPolygon(new SolidBrush(Color.White), blockPoints);
            }
        }

        /// <summary>
        /// Berekent op basis van twee hoofdkleuren de tussenliggende gradientkleuren.
        /// </summary>
        /// <param name="c1">Eerste hoofdkleur.</param>
        /// <param name="c2">Tweede hoofdkleur.</param>
        /// <param name="depth">Aantal tussenkleuren min twee.</param>
        /// <returns>Een array met de hoofdkleuren en tussenkleuren (gesorteerd vanaf de eerste hoofdkleur).</returns>
        private Pen[] ComputeGradientColors(Color c1, Color c2, int depth)
        {
            //Init.
            Pen[] pens = new Pen[depth];
            int i, r, g, b;

            //Bereken gradientkleuren.
            for (i = 0; i < depth - 1; i++)
            {
                r = ((c2.R - c1.R) / depth) * i + c1.R; g = ((c2.G - c1.G) / depth) * i + c1.G;
                b = ((c2.B - c1.B) / depth) * i + c1.B;
                pens[i] = new Pen(Color.FromArgb(r, g, b));
            }

            if (pens.Length == 0) System.Diagnostics.Debugger.Break();

            pens[depth - 1] = new Pen(c2);


            //Finish.
            return pens;
        }
    }
}
