﻿/*** Name:      Block.cs
 *** Created:   10/25/2008, by: Hilkens, RGM
 *** Modified:  12/1/2008, by: Hilkens, RGM
 *** Source:    Definieert een blok.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace Etaris.Fields
{
    /// <summary>
    /// Definieert een blok. Een blok is altijd onderdeel van een speelstuk.
    /// </summary>
    public class Block : BaseBlock
    {
        /// <summary>
        /// Speelstuk waarvan dit blok onderdeel uitmaakt.
        /// </summary>
        internal Piece Piece;

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        /// <param name="piece">Een speelstuk.</param>
        /// <param name="left">X-positie in het speelveld tellend vanaf 0 (van links). De eenheid is een blokbreedte.</param>
        /// <param name="top">Y-positie in het speelveld, tellend vanaf 0 (van boven). De eenheid is een blokhoogte.</param>
        internal Block(Piece piece, int left, int top)
            : base(left, top)
        {
            this.Piece = piece;
            PlayingField field = piece.Field;

            if (field.Grid[left, top] != null) throw new Exception("Improper block position.");
            field.RowItemCount[top]++;
            field.Grid[left, top] = this;
        }

        /// <summary>
        /// Controleert of een blok kan bewegen in een bepaalde richting.
        /// </summary>
        /// <param name="direction">De richting waarvan gecontroleerd moet worden of een blok erin kan bewegen</param>
        /// <returns><i>true</i> indien het blok in de aangegeven richting kan bewegen, anders <i>false</i>.</returns>
        public bool CanMove(Direction direction)
        {
            //Init.
            Piece p = null;

            switch (direction)
            {
                case Direction.Down:
                    if (Top == Piece.Field.Height - 1) return false;
                    p = Piece.Field.GetPiece(Left, Top + 1);
                    break;
                case Direction.Left:
                    if (Left == 0) return false;
                    p = Piece.Field.GetPiece(Left - 1, Top);
                    break;
                case Direction.Right:
                    if (Left == Piece.Field.Width - 1) return false;
                    p = Piece.Field.GetPiece(Left + 1, Top);
                    break;
                case Direction.Up:
                    if (Top == 0) return false;
                    p = Piece.Field.GetPiece(Left, Top - 1);
                    break;
            }

            //Finish.
            if (p == null) return true;
            return (p == Piece);
        }

        /// <summary>
        /// Wijzigt de positie en werkt het bijbehorende grid bij.
        /// </summary>
        /// <param name="newLeft">X-positie in het speelveld tellend vanaf 0 (van links). De eenheid is een blokbreedte.</param>
        /// <param name="newTop">Y-positie in het speelveld tellend vanaf 0 (van boven). De eenheid is een blokhoogte.</param>
        /// <returns><i>true</i> als de wijziging met succes voltooid is, anderzijds <i>false</i>.</returns>
        internal bool SetNewPosition(int newLeft, int newTop)
        {
            //Init.
            PlayingField field = Piece.Field;

            //Controleer of verandering is toegestaan.
            if (Piece.Field.Grid[newLeft, newTop] != null && Piece.Field.Grid[newLeft, newTop].Piece != Piece) return false;

            //Werk RowItemCount bij.
            if (newTop != Top) { Piece.Field.RowItemCount[Top]--; Piece.Field.RowItemCount[newTop]++; }

            //Werk Grid bij. Onderstaande controle is nodig omdat een verplaatst blok van hetzelfde speelstuk
            //het grid tijdelijk in een inconsistente staat zet.
            if (Piece.Field.Grid[Left, Top] == this) Piece.Field.Grid[Left, Top] = null;
            this.Left = newLeft; this.Top = newTop;
            Piece.Field.Grid[Left, Top] = this;

            //Finish.
            return true;
        }

        /// <summary>
        /// Controleert of een zijde <i>side</i> een rand van het speelstuk is.
        /// </summary>
        /// <remarks>Override de functie in <i>BaseBlock</i></remarks>
        /// <param name="piece">Een basisspeelstuk, wordt niet gebruikt</param>
        /// <param name="side">Een zijde.</param>
        /// <returns><i>true</i> als Side een rand van het speelstuk is, anderzijds <i>false</i>.</returns>
        internal override bool IsBorder<T>(BasePiece<T> piece, Sides side)
        {
            //Init.
            int dx = 0, dy = 0;

            switch (side)
            {
                case Sides.Top:
                    dy = -1;
                    break;
                case Sides.Right:
                    dx = 1;
                    break;
                case Sides.Bottom:
                    dy = 1;
                    break;
                case Sides.Left:
                    dx = -1;
                    break;
            }

            //Finish.
            return (Piece.Field.GetPiece(this.Left + dx, this.Top + dy) != Piece);
        }
    }
}