﻿/*** Name:      Border.cs
 *** Created:   11/15/2008, by: Hilkens, RGM
 *** Modified:  12/14/2008, by: Hilkens, RGM
 *** Source:    Bereken randen van een speelstuk.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Diagnostics;

namespace Etaris.Fields
{
    /// <summary>
    /// Berekent de GDI+ coördinaten van de randen van een speelstuk.
    /// </summary>
    /// <remarks>
    /// Een rand is een set lijnstukken welke gesorteerd zijn en een oppervlakte omsluiten.
    /// Het tweede coördinaat van het laatste lijnstuk komt altijd overeen met het eerste coördinaat
    /// van het eerste lijnstuk.
    /// </remarks>
    internal class Border<T> where T : BaseBlock
    {
        /// <summary>
        /// Vectorvergelijker, wordt gebruikt bij de zoeken en sorteren.
        /// </summary>
        IComparer<Vector> c = new VectorComparor();

        /// <summary>
        /// De begin- en eindpunten van lijnstukken.
        /// </summary>
        Vector[] points;

        /// <summary>
        /// De lijnstukken die samen de randen vormen.
        /// </summary>
        List<Line> lines = new List<Line>();

        /// <summary>
        /// Een rand.
        /// </summary>
        List<Line> linesSorted = new List<Line>();

        /// <summary>
        /// Alle randen (ook binnenranden).
        /// </summary>
        List<List<Line>> linesSortedAll = new List<List<Line>>();

        /// <summary>
        /// De grootte van een blok.
        /// </summary>
        int size;

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        internal Border()
        { }

        /// <summary>
        /// Berekent van een speelstuk alle randen (voor gradientkleuring).
        /// </summary>
        /// <param name="piece">Een speelstuk.</param>
        /// <param name="blockSize">Lengte/breedte van een blok.</param>
        /// <param name="depth">Diepte van shading.</param>
        /// <returns>Lijst met gesorteerde coördinaten Point[i][j] van een rand. Hierbij is i=0 de buitenste rand en i=1 de tweede etc...</returns>
        internal List<Point[][]> ProcessPiece(BasePiece<T> piece, int blockSize, int depth)
        {
            //Init
            int i, j;
            List<Point[][]> points;
            size = blockSize;

            lines.Clear();
            linesSorted.Clear();

            ExtractOuterLines(piece);
            ExtractPoints();
            SortLinesAll();

            points = new List<Point[][]>();

            for (i = 0; i < linesSortedAll.Count; i++)
            {
                linesSorted = linesSortedAll[i];
                points.Add(new Point[depth][]);
                points[i][0] = ComputeGDILines(linesSorted);

                for (j = 1; j < depth; j++)
                {
                    ShrinkBorder(linesSorted); //merk op dat MakeShading linesSorted aanpast!
                    points[i][j] = ComputeGDILines(linesSorted);
                }
            }

            //Finish.
            return points;
        }

        /// <summary>
        /// Verkleint een rand.
        /// </summary>
        /// <param name="lines">Een rand.</param>
        private void ShrinkBorder(List<Line> lines)
        {
            //Init
            int i, c = lines.Count;
            int prev, next;

            //Doorloop alle lijnstukken en verplaats ze naar 'binnen'.
            for (i = 0; i < c; i++)
            {
                if (lines[i].Orientation == Sides.Top) { lines[i].p1.Y++; lines[i].p2.Y++; }
                if (lines[i].Orientation == Sides.Right) { lines[i].p1.X--; lines[i].p2.X--; }
                if (lines[i].Orientation == Sides.Bottom) { lines[i].p1.Y--; lines[i].p2.Y--; }
                if (lines[i].Orientation == Sides.Left) { lines[i].p1.X++; lines[i].p2.X++; }

                if (i == 0) prev = c - 1; else prev = i - 1; //Gaat uit van omsloten oppervlak.
                if (i == (c - 1)) next = 0; else next = i + 1;
                lines[prev].p2 = lines[i].p1;
                lines[next].p1 = lines[i].p2;
            }
        }

        /// <summary>
        /// Berekent GDI coördinaten (afronding op pixels) van gesorteerde lijnstukken, cyclisch.
        /// </summary>
        /// <param name="lines">Een rand.</param>
        /// <returns>Een array met GDI coördinaten.</returns>
        private Point[] ComputeGDILines(List<Line> lines)
        {
            //Init.
            int i;
            Point[] points = new Point[lines.Count];
            Line line = null;
            Point previousP2 = new Point();
            Point p1 = new Point();
            Point p2 = new Point();

            //Doorloop alle lijnstukken.
            for (i = 0; i < lines.Count; i++)
            {
                line = lines[i];
                p1 = line.p1.GetPoint(); p2 = line.p2.GetPoint();

                if (line.Orientation == Sides.Top || line.Orientation == Sides.Bottom)
                {
                    if (line.p1.X > line.p2.X) p1.X -= 1; else p2.X -= 1;           //Bepaal X waarde
                    if (line.Orientation == Sides.Bottom) { p1.Y -= 1; p2.Y -= 1; } //Bepaal Y waarde
                    if (i != 0 && previousP2 != p1) p1.X = previousP2.X;            //Correctie
                }
                else
                {
                    if (line.p1.Y > line.p2.Y) p1.Y -= 1; else p2.Y -= 1;           //Bepaal X waarde
                    if (line.Orientation == Sides.Right) { p1.X -= 1; p2.X -= 1; }  //Bepaal Y waarde
                    if (i != 0 && previousP2 != p1) p1.Y = previousP2.Y;            //Correctie
                }

                previousP2 = p2;
                points[i] = p1;
            }

            //Correctie voor het eerste coördinaat.
            if (line.Orientation == Sides.Top || line.Orientation == Sides.Bottom) points[0].Y = p1.Y;
            else points[0].X = p1.X;

            //Finish.
            return points;
        }

        /// <summary>
        /// Voegt een lijnstuk in volgorde toe aan het eindresultaat.
        /// </summary>
        /// <remarks>Deze functie wordt door <i>SortLines</i>() aangeroepen.</remarks>
        /// <param name="line">Een nieuw lijnstuk.</param>
        private void AddToFinal(Line line)
        {
            //Init.
            int l; List<Line> list = linesSorted;

            if (list.Count == 0) list.Add(line);
            else
            {
                l = linesSorted.Count - 1;

                if ((list[l].p1.X == list[l].p2.X && list[l].p2.X == line.p2.X) ||
                    (list[l].p1.Y == list[l].p2.Y && list[l].p2.Y == line.p2.Y))
                { list[l].p2 = line.p2; }
                else list.Add(line);
            }
        }

        /// <summary>
        /// Sorteert en zoekt alle mogelijke randen.
        /// </summary>
        private void SortLinesAll()
        {
            //Init.
            Line line = null;
            int i;

            //Verwerkt alle lijnstukken.
            while (true)
            {
                for (i = 0; i < lines.Count; i++)
                {
                    if (!lines[i].Processed)
                    {
                        line = lines[i];
                        linesSorted = new List<Line>();
                        linesSortedAll.Add(linesSorted);
                        break;
                    }
                }
                if (i == lines.Count) break;
                SortLines(line.p1);
            }
        }

        /// <summary>
        /// Vindt een geschikt vervolg lijnstuk. Startende op v1.
        /// </summary>
        /// <param name="v1">Een vector. Vector moet tot een lijnstuk behoren.</param>
        /// <returns>Een index in <i>points</i>.</returns>
        private int FindVector(Vector v1)
        {
            //Init.
            int i = Array.BinarySearch<Vector>(points, v1, c);
            int answer = -1, original = i;
            Line oldLine = v1.line, newLine;
            bool goBack = true;

            while (true)
            {
                if (Vector.IsEqual(points[i], v1))
                {
                    //Mogelijk antwoord.
                    newLine = points[i].line;
                    if (newLine != oldLine)
                    {
                        if (oldLine.Block == newLine.Block) answer = i;
                        else if (answer == -1) answer = i;
                    }
                }
                else
                {
                    if (goBack == false) break; else { goBack = false; i = original; }
                }

                //Bepaal volgende item.
                if (i == 0 && goBack) { goBack = false; i = original; }
                if (i == points.Length - 1 && !goBack) break;
                if (goBack) i--; else i++;
            }

            //Finish.
            return answer;
        }


        /// <summary>
        /// Construeert een rand door de lijnstukken te sorteren.
        /// </summary>
        private void SortLines(Vector start)
        {
            //Init.
            int i;
            List<Line> border = new List<Line>();
            Line l2;
            Vector v1, v2, v1b; v1 = start;

            while (true)
            {
                i = FindVector(v1);
                if (i < 0) throw new Exception();

                //Verwerk resultaat
                v1b = points[i];
                l2 = points[i].line;
                if (l2.Processed) break;

                //Vraag bijbehorende nieuwe punt op
                if (v1b.Equals(l2.p1)) v2 = l2.p2; else v2 = l2.p1;
                AddToFinal(Line.GetLine(v1b.X, v1b.Y, v2.X, v2.Y, l2.Orientation, null));

                //Finish.
                l2.Processed = true;
                v1 = v2;
            }
        }

        /// <summary>
        /// Voegt alle coördinaten van alle lijnstukken toe aan een geordende lijst.
        /// </summary>
        private void ExtractPoints()
        {
            //Init.
            int i, j = 0;
            points = new Vector[lines.Count * 2];

            //Doorloop alle lijnstukken.
            for (i = 0; i < lines.Count; i++)
            {
                points[j] = lines[i].p1;
                points[j + 1] = lines[i].p2;
                j += 2;
            }

            //Sorteer coördinaten, zo kan later BinarySearch gebruikt worden.
            Array.Sort<Vector>(points, c);
        }

        /// <summary>
        /// Berekent alle buitenste lijnstukken van een speelstuk.
        /// </summary>
        /// <param name="piece">Een speelstuk.</param>
        private void ExtractOuterLines(BasePiece<T> piece)
        {
            //Init.
            int i;
            T block;

            //Doorloop alle blokken van een speelstuk.
            for (i = 0; i < piece.Blocks.Count; i++)
            {
                block = piece.Blocks[i];
                if (block.IsBorder<T>(piece, Sides.Top))
                    lines.Add(Line.GetLine(block.Left * size, block.Top * size, block.Left * size + size, block.Top * size, Sides.Top, block));

                if (block.IsBorder<T>(piece, Sides.Right))
                    lines.Add(Line.GetLine(block.Left * size + size, block.Top * size, block.Left * size + size, block.Top * size + size, Sides.Right, block));

                if (block.IsBorder<T>(piece, Sides.Bottom))
                    lines.Add(Line.GetLine(block.Left * size, block.Top * size + size, block.Left * size + size, block.Top * size + size, Sides.Bottom, block));

                if (block.IsBorder<T>(piece, Sides.Left))
                    lines.Add(Line.GetLine(block.Left * size, block.Top * size, block.Left * size, block.Top * size + size, Sides.Left, block));
            }
        }

        /// <summary>
        /// Representatie van een verticaal of horizontaal lijnstuk. Deze klasse is alleen nodig voor de <i>Border</i> klasse.
        /// </summary>
        private class Line
        {
            /// <summary>
            /// Punt 1.
            /// </summary>
            internal Vector p1;

            /// <summary>
            /// Punt 2.
            /// </summary>
            internal Vector p2;

            /// <summary>
            /// Orientatie van het lijnstuk.
            /// </summary>
            internal Sides Orientation;

            /// <summary>
            /// Bijbehorend blok.
            /// </summary>
            internal BaseBlock Block;

            /// <summary>
            /// Geeft aan of dit lijnstuk verwerkt is door een willekeurig algoritme.
            /// </summary>
            internal bool Processed = false;

            /// <summary>
            /// Constructor. Initialiseert de klasse.
            /// </summary>
            private Line()
            { }

            /// <summary>
            /// Geeft een <i>Line</i> object op basis van twee coördinaten en een orientatie.
            /// </summary>
            /// <param name="x1">X-coördinaat van punt 1.</param>
            /// <param name="y1">Y-coördinaat van punt 1.</param>
            /// <param name="x2">X-coördinaat van punt 2.</param>
            /// <param name="y2">Y-coördinaat van punt 2.</param>
            /// <param name="Orientation">Orientatie van het lijnstuk (zie aparte documentatie).</param>
            /// <param name="block">Het blok waarvan deze lijn een deel van de rand is.</param>
            /// <returns>Een <i>Line</i> object.</returns>
            internal static Line GetLine(int x1, int y1, int x2, int y2, Sides Orientation, BaseBlock block)
            {
                Line line = new Line();
                line.p1.X = x1; line.p1.Y = y1;
                line.p1.line = line;
                line.p2.X = x2; line.p2.Y = y2;
                line.p2.line = line;
                line.Block = block;

                line.Orientation = Orientation;
                return line;
            }
        }

        /// <summary>
        /// Representatie van een vector. Deze structuur is alleen nodig voor de <i>Border</i> klasse.
        /// </summary>
        private struct Vector
        {
            /// <summary>
            /// X-coördinaat van de vector.
            /// </summary>
            internal int X;

            /// <summary>
            /// Y-coördinaat van de vector.
            /// </summary>
            internal int Y;

            /// <summary>
            /// Bijbehorend lijnstuk.
            /// </summary>
            internal Line line;

            /// <summary>
            /// Maakt een GDI Point object van deze vector.
            /// </summary>
            /// <returns></returns>
            internal Point GetPoint()
            {
                return new Point(X, Y);
            }

            /// <summary>
            /// Is-gelijk operator.
            /// </summary>
            /// <param name="v1">Vector 1.</param>
            /// <param name="v2">Vector 2.</param>
            /// <returns><i>true</i> als <i>v1</i> en <i>v2</i> identiek zijn, anderzijds <i>false</i>.</returns>
            internal static bool IsEqual(Vector v1, Vector v2)
            {
                return (v1.X == v2.X && v1.Y == v2.Y);
            }

            /// <summary>
            /// Geeft een string-representatie van deze structuur.
            /// </summary>
            /// <returns>Een string.</returns>
            public override string ToString()
            {
                return "{" + X + ", " + Y + "}";
            }
        }

        /// <summary>
        /// Vergelijkingsklasse voor vectoren. Deze is alleen nodig voor de <i>Border</i> klasse.
        /// </summary>
        private class VectorComparor : IComparer<Vector>
        {
            /// <summary>
            /// Vergelijkt twee vectoren.
            /// </summary>
            /// <param name="p1">Vector 1.</param>
            /// <param name="p2">Vector 2.</param>
            /// <returns>Zie .NET documentatie.</returns>
            int IComparer<Vector>.Compare(Vector p1, Vector p2)
            {
                if (p1.X < p2.X) return -1;
                else if (p1.X > p2.X) return 1;
                else
                {
                    if (p1.Y < p2.Y) return -1;
                    else if (p1.Y > p2.Y) return 1;
                    else return 0;
                }
            }
        }
    }
}
