﻿/*** Name:      Piece.Partial.cs
 *** Created:   12/1/2008, by: Hilkens, RGM
 *** Modified:  12/13/2008, by: van Vliembergen, RWL
 *** Source:    Representatie van een massamiddelpunt en een boundingbox (nodig voor rotatie).
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;

namespace Etaris.Fields
{
    partial class Piece
    {
        /// <summary>
        /// Representatie van een massamiddelpunt.
        /// </summary>
        private struct MassCentre
        {
            /// <summary>
            /// <i>true</i> als het massamiddelpunt is toegekend, anderzijds <i>false</i>.
            /// </summary>
            internal bool IsCalculated;

            /// <summary>
            /// X-coördinaat van het massamiddelpunt.
            /// </summary>
            internal int X;

            /// <summary>
            /// Y-coördinaat van het massamiddelpunt.
            /// </summary>
            internal int Y;

            /// <summary>
            /// String-representatie van deze klasse in speelveldcoördinaten.
            /// </summary>
            /// <returns>Een string.</returns>
            public override string ToString()
            {
                return "(" + (X / 2) + ", " + (Y / 2) + ")";
            }
        }

        /// <summary>
        /// Representeert een omsloten rechthoek voor een speelstuk. Dit is nodig voor controle op botsingen bij het roteren van een speelstuk.
        /// Merk op dat rotatie nog steeds om het massamiddelpunt gaat.
        /// </summary>
        /// <remarks>(x1, y1) is altijd het kleinste, (x2, y2) is altijd het grootste coördinatenpaar (validiteit).</remarks>
        private struct BoundingBox
        {
            /// <summary>
            /// De horizontale coördinaat van de linkerbovenhoek van het bijbehorende speelstuk
            /// </summary>
            internal int x1;

            /// <summary>
            /// De verticale coördinaat van de linkerbovenhoek van het bijbehorende speelstuk
            /// </summary>
            internal int y1;

            /// <summary>
            /// De horizontale coördinaat van de rechteronderhoek van het bijbehorende speelstuk
            /// </summary>
            internal int x2;

            /// <summary>
            /// De verticale coördinaat van de rechteronderhoek van het bijbehorende speelstuk
            /// </summary>
            internal int y2;

            /// <summary>
            /// Maakt een (gedraaide) omsloten rechthoek valide.
            /// </summary>
            internal void MakeValid()
            {
                int temp;
                if (x1 > x2) { temp = x1; x1 = x2; x2 = temp; }
                if (y1 > y2) { temp = y1; y1 = y2; y2 = temp; }
            }

            /// <summary>
            /// Werkt de huidige omsloten rechthoek bij met de vereniging van de gegeven omsloten rechthoeken.
            /// </summary>
            /// <param name="box">Een boundingbox.</param>
            internal void Union(BoundingBox box)
            {
                if (box.x1 < x1) x1 = box.x1;
                if (box.y1 < y1) y1 = box.y1;
                if (box.x2 > x2) x2 = box.x2;
                if (box.y2 > y2) y2 = box.y2;
            }

            /// <summary>
            /// Berekent de omsloten rechthoek van het gegeven speelstuk.
            /// </summary>
            /// <param name="piece">Een speelstuk.</param>
            internal void Calculate(Piece piece)
            {
                //Controleer Blocks.
                if (piece.blocks.Count == 0)
                { y1 = 0; y1 = 0; y2 = 0; x2 = 0; return; }

                x1 = piece.blocks[0].Left;
                y1 = piece.blocks[0].Top;

                x2 = x1;
                y2 = y1;

                //Doorloop alle blokken.
                for (int i = 1; i < piece.blocks.Count; i++)
                {
                    if (piece.blocks[i].Left < x1)
                        x1 = piece.blocks[i].Left;
                    if (piece.blocks[i].Left > x2)
                        x2 = piece.blocks[i].Left;
                    if (piece.blocks[i].Top < y1)
                        y1 = piece.blocks[i].Top;
                    if (piece.blocks[i].Top > y2)
                        y2 = piece.blocks[i].Top;
                }
            }

            /// <summary>
            /// Berekent of er zich in de omsloten rechthoek speelstukken bevinden, <i>piece</i> wordt niet meegerekend.
            /// </summary>
            /// <param name="piece">Een speelstuk.</param>
            /// <returns><i>true</i> als er zich speelstukken anders dan <i>piece</i> zich in deze omsloten rechthoek bevinden, anderzijds <i>false</i>.</returns>
            internal bool IsGridFree(Piece piece)
            {
                //Init.
                Block[,] grid = piece.field.Grid;
                int widthUp = grid.GetUpperBound(0);
                int heightUp = grid.GetUpperBound(1);

                //Controleer of deze boundingbox zich in grid bevindt. 
                if (x1 < 0 || x2 < 0 || x1 > widthUp || x2 > widthUp ||
                    y1 < 0 || y2 < 0 || y1 > heightUp || y2 > heightUp) return false;

                //Doorloop alle blokken.
                for (int i = x1; i <= x2; i++)
                    for (int j = y1; j <= y2; j++)
                        if (grid[i, j] != null && grid[i, j].Piece != piece) return false;

                //Finish.
                return true;
            }
        }
    }
}
