﻿/*** Name:      PlayingField.cs
 *** Created:   11/6/2008, by: Hilkens, RGM
 *** Modified:  12/14/2008, by: Hilkens, RGM
 *** Source:    Definieert een speelveld bestaande uit speelstukken.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Etaris.Fields
{
    /// <summary>
    /// Definieert een speelveld bestaande uit speelstukken. Deze klasse bevat geen spellogica.
    /// </summary>
    /// <remarks>Coördinaten zijn altijd gehele getallen paren (X, Y).</remarks>
    public class PlayingField
    {
        #region Velden
        /// <summary>
        /// De lijst met speelstukken die zich in dit speelveld bevinden.
        /// </summary>
        LinkedList<Piece> pieces;

        /// <summary>
        /// Lengte/breedte van een blok.
        /// </summary>
        int blockSize;

        /// <summary>
        /// Diepte van een gradient (schaduwrand van speelstukken).
        /// </summary>
        int shadingSize;

        /// <summary>
        /// De grafische context. Op deze context wordt getekend.
        /// </summary>
        Graphics graphics;

        /// <summary>
        /// Het rooster met de blokken.
        /// </summary>
        Block[,] grid;

        /// <summary>
        /// Aantal blokken in alle kolommen <i>j</i> van een rij <i>index</i> in <i>Grid[j, index]</i>.
        /// </summary>
        int[] rowItemCount;

        /// <summary>
        /// Breedte in blokken van dit speelveld
        /// </summary>
        int width;

        /// <summary>
        /// Hoogte in blokken van dit speelveld.
        /// </summary>
        int height;
        #endregion

        #region "Properties"
        /// <summary>
        /// De speelstukken.
        /// </summary>
        public LinkedList<Piece> Pieces
        {
            get { return pieces; }
            set { pieces = value; }
        }

        /// <summary>
        /// Lengte/breedte van een blok.
        /// </summary>
        public int BlockSize
        {
            get { return blockSize; }
            set
            {
                blockSize = value;
                foreach (Piece piece in pieces)
                    piece.SetPaintingSizes(blockSize, shadingSize);
            }
        }

        /// <summary>
        /// Diepte van een gradient (schaduwrand van speelstukken).
        /// </summary>
        public int ShadingSize
        {
            get { return shadingSize; }
            set
            {
                shadingSize = value;
                foreach (Piece piece in pieces)
                    piece.SetPaintingSizes(blockSize, shadingSize);
            }
        }

        /// <summary>
        /// De grafische context. Op deze context wordt getekend.
        /// </summary>
        public Graphics GraphicsContext
        {
            internal get { return graphics; }
            set { graphics = value; }
        }

        /// <summary>
        /// Het rooster met de blokken.
        /// </summary>
        internal Block[,] Grid
        {
            get { return grid; }
            set { grid = value; }
        }

        /// <summary>
        /// Aantal blokken op een regel.
        /// </summary>
        internal int[] RowItemCount
        {
            get { return rowItemCount; }
            set { rowItemCount = value; }
        }

        /// <summary>
        /// Breedte van het speelveld in blokken.
        /// </summary>
        internal int Width
        {
            get { return width; }
        }

        /// <summary>
        /// Hoogte van het speelveld in blokken.
        /// </summary>
        internal int Height
        {
            get { return height; }
        }
        #endregion

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        /// <param name="blockSize">Lengte/breedte van een blok.</param>
        /// <param name="shadingSize">Diepte van een gradient (schaduwrand van speelstukken).</param>
        /// <param name="width">Breedte van het speelveld in blokken.</param>
        /// <param name="height">Hoogte van het speelveld in blokken.</param>
        public PlayingField(int blockSize, int shadingSize, int width, int height)
        {
            //Init.
            this.pieces = new LinkedList<Piece>();
            this.blockSize = blockSize;
            this.shadingSize = shadingSize;
            this.width = width;
            this.height = height;
            grid = new Block[width, height];
            rowItemCount = new int[height];  //Soortgelijk voor de rowItemCount.

            //Het Grid initialiseren.
            for (int i = 0; i < width; i++)
                for (int j = 0; j < height; j++)
                    grid[i, j] = null;

            //rowItemCount initialiseren.
            for (int i = 0; i < height; i++)
                rowItemCount[i] = 0;
        }

        /// <summary>
        /// Tekent alle blokken in een speelveld.
        /// </summary>
        public void Paint()
        {
            //Init.
            Region r = new Region(); r.MakeInfinite();

            graphics.FillRegion(Brushes.White, r); //Verwijder de oude tekening

            //Finish.
            foreach (Piece pL in pieces)
                pL.Paint(graphics);
        }

        /// <summary>
        /// Verwijdert alle blokken op de gegeven rijen.
        /// </summary>
        /// <param name="rows">Y-posities van de te verwijderen rijen in het speelveld, tellend vanaf 0 (van boven).</param>
        public void RemoveRows(int[] rows)
        {
            //Init.
            int counter = 0;
            Piece[] A = new Piece[width * rows.Length];
            List<Block>[] B = new List<Block>[width * rows.Length];
            Piece p;

            //Initialiseer Tag in de speelstukken in rows.
            foreach (int j in rows)
                for (int i = 0; i < width; i++)
                    grid[i, j].Piece.Tag = -1; //Dit speelstuk hebben we nog niet aan A en B toegevoegd

            //Zoek de blokken die verwijderd moeten worden, en de speelstukken waarin ze zitten.
            foreach (int j in rows)
            {
                for (int i = 0; i < width; i++)
                {
                    p = GetPiece(i, j);
                    if (p == null) continue;

                    if (p.Tag == -1) //Eerste voorkomen van dit speelstuk in deze rij(en)
                    {
                        p.Tag = counter;
                        B[counter] = new List<Block>();
                        A[counter++] = p;
                    }
                    B[p.Tag].Add(grid[i, j]);
                }
            }

            //Verwijder de blokken, en splits of verwijder het speelstuk indien nodig.
            for (int i = 0; i < counter; i++)
                A[i].RemoveBlocks(B[i]);
        }

        /// <summary>
        /// Controleert of een rij vol is.
        /// </summary>
        /// <param name="i">Y-positie in het speelveld, tellend vanaf 0 (van boven).</param>
        /// <returns><i>true</i> als de rij helemaal vol is, anders <i>false</i>.</returns>
        public bool IsRowFull(int i)
        {
            return (rowItemCount[i] == width);
        }

        /// <summary>
        /// Controleert of een rij leeg is.
        /// </summary>
        /// <param name="i">Y-positie in het speelveld, tellend vanaf 0 (van boven).</param>
        /// <returns><i>true</i> als de rij helemaal leeg is, anders <i>false</i>.</returns>
        public bool IsRowEmpty(int i)
        {
            return (rowItemCount[i] == 0);
        }

        /// <summary>
        /// Zoekt een speelstuk op de gegeven positie (<i>left</i>, <i>top</i>).
        /// </summary>
        /// <param name="left">X-positie in het speelveld, tellend vanaf 0 (van links).</param>
        /// <param name="top">Y-positie in het speelveld, tellend vanaf 0 (van boven).</param>
        /// <returns>Het gevonden speelstuk.</returns>
        public Piece GetPiece(int left, int top)
        {
            if ((left < 0) || (left >= width))
                return null;
            if ((top < 0) || (top >= height))
                return null;
            if (grid[left, top] == null)
                return null;
            return grid[left, top].Piece;
        }
    }
}
